import logging
import os
import re
import textwrap
import subprocess
import tempfile
import munch
import shlex
import fnmatch
import argparse
import pycurl
import sys

from .errors import RpkgError

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

log = logging.getLogger("__main__")

QUIET = 0
MODEST = 1
VERBOSE = 2

MACRO_HELPER_DIR = '/usr/lib/rpkg.macros.d/helpers/'
UPLOAD_EXT = ['tar', 'gz', 'bz2', 'lzma', 'xz', 'Z', 'zip', 'tff',
              'bin', 'tbz', 'tbz2', 'tgz', 'tlz', 'txz', 'pdf', 'rpm',
              'jar', 'war', 'db', 'cpio', 'jisp', 'egg', 'gem', 'spkg',
              'oxt', 'xpi', 'crate']

def changelog_entry(commit_message):
    entry = '- ' + commit_message.split('\n')[0] if commit_message else ''
    return '\n'.join(textwrap.wrap(re.sub(r'(\s+)%([^%])', r'\1%%\2', entry), 80))


def run(cmd, shell=False, env=None, cwd=None, capture=True, capture_stderr=False, throw=True):
    if capture:
        stdout = subprocess.PIPE
    else:
        stdout = None

    if capture_stderr:
        stderr = subprocess.PIPE
    else:
        stderr = None

    log.debug('Running: {0}'.format(
        cmd_repr(cmd) if not shell else cmd))

    process = subprocess.Popen(
        cmd, stdout=stdout, stderr=stderr, shell=shell, env=env, cwd=cwd)

    (out, err) = process.communicate()

    err = err.decode('utf-8').strip('\n') if err else ''

    if process.returncode != 0 and throw:
        raise RpkgError(err)

    out = out.decode('utf-8').strip('\n') if out else ''

    return munch.Munch(out=out,
                       err=err,
                       rc=process.returncode)


def edit(editor, text):
    tmp_f = tempfile.NamedTemporaryFile('w', delete=False)
    tmp_f.write(text)
    tmp_f.close()
    subprocess.check_call(editor.split()+[tmp_f.name])
    tmp_f = open(tmp_f.name, 'r')
    result = tmp_f.read()
    tmp_f.close()
    os.unlink(tmp_f.name)
    return result


def parse_config_rpmdefines(rpmdefines_str):
    return [item.strip('"\'') for item in re.findall(r'(\'[^\']*\'|"[^"]*")', rpmdefines_str)]


def parse_config_list(path_list_str):
    return [item.strip('"\'') for item in path_list_str.split()]


def cmd_repr(cmd):
    quoted_items = []

    for i in range(len(cmd)):
        quoted_items.append(shlex.quote(cmd[i]))

    return ' '.join(quoted_items)


def has_scheme(path):
    return urlparse(path).scheme != ''


def macro_helper_cmd(*args):
    return [MACRO_HELPER_DIR+args[0], *args[1:]]


def extract_srpm(srpm_path, destdir):
    cmd = 'rpm2cpio {0} | cpio -idm --quiet'.format(shlex.quote(srpm_path))
    run(cmd, cwd=destdir, shell=True)


def is_upload_item(item):
    for upload_ext in UPLOAD_EXT:
        if item.endswith('.'+upload_ext):
            return True
    return False


def is_reserved_item(item, script_name='rpkg'):
    for pattern in [script_name+'.conf', '.'+script_name+'*', 'sources', 'README*', '.git*']:
        if fnmatch.fnmatch(item, pattern):
            return True
    return False


class DoNothing(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        pass


def print_network_progress(to_download, downloaded, to_upload, uploaded):
    if not sys.stdout.isatty():
        # Don't print progress if not outputting into TTY. The progress
        # output is not useful in logs.
        return

    if to_download > 0:
        done = downloaded / to_download
    elif to_upload > 0:
        done = uploaded / to_upload
    else:
        return

    done_chars = int(done * 72)
    remain_chars = 72 - done_chars
    done = int(done * 1000) / 10.0

    p = "\r%s%s %s%%" % ("#" * done_chars, " " * remain_chars, done)
    sys.stdout.write(p)
    sys.stdout.flush()


def download_file(download_url, outfile, max_redirs=1, ca_cert=None):
    with open(outfile, 'wb') as f:
        c = pycurl.Curl()
        c.setopt(pycurl.URL, str(download_url))
        c.setopt(pycurl.HTTPHEADER, ['Pragma:'])
        c.setopt(pycurl.NOPROGRESS, False)
        c.setopt(pycurl.PROGRESSFUNCTION, print_network_progress)
        c.setopt(pycurl.OPT_FILETIME, True)
        c.setopt(pycurl.WRITEDATA, f)
        c.setopt(pycurl.LOW_SPEED_LIMIT, 1000)
        c.setopt(pycurl.LOW_SPEED_TIME, 300)
        c.setopt(pycurl.FOLLOWLOCATION, True)
        c.setopt(pycurl.MAXREDIRS, max_redirs)

        if ca_cert:
            c.setopt(pycurl.CAINFO, ca_cert)

        try:
            c.perform()
            tstamp = c.getinfo(pycurl.INFO_FILETIME)
            status = c.getinfo(pycurl.RESPONSE_CODE)

        except Exception as e:
            raise RpkgError(e)

        finally:
            c.close()

    # Render new-line after displaying the progress
    sys.stdout.write('\n')
    sys.stdout.flush()

    if tstamp:
        os.utime(outfile, (tstamp, tstamp))

    return status
